from abc import ABC
from bc_time.api.api import Api

class Base(ABC):
    api = None
    content_type_id = None

    def __init__(self, api: Api=None) -> None:
        self.api = Api() if not api else api


from bc_time.api.objects.object_base import ObjectBase
from bc_time.api.api import Api
from bc_time.api.enumerators.content_type import ContentType

class Employees(ObjectBase):
    def __init__(self, api: Api=None) -> None:
        super().__init__(api)
        self.content_type_id = ContentType.employee



from os import urandom as os_urandom
from base64 import b64encode as base64_encode, b64decode as base64_decode
from random import choice as random_choice
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives import padding

class Crypt:
    IV_START_VALUES = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H'] # a = 1, b = 2..., A = 9, B = 10..., and etc.
    KEY_LENGTH = 32 # Not used in this class (only in BC Time), but leaving it here for information sake.
    IV_LENGTH = 16

    key = None
    data = None

    def __init__(self, key=None, data=None) -> None:
        self.key = key
        self.data = data

    def encrypt(self) -> str:
        if not self.__validate_data():
            return None
        iv, iv_start, iv_end = self.__get_iv_and_start_and_end()
        padder = padding.PKCS7(128).padder()
        data_to_encrypt = padder.update(str.encode(self.data)) + padder.finalize()
        key = str.encode(self.key)
        iv = str.encode(iv)
        cipher = Cipher(algorithms.AES(key), modes.CBC(iv))
        encryptor = cipher.encryptor()
        encrypted_data = encryptor.update(data_to_encrypt) + encryptor.finalize()
        encrypted_data = str(base64_encode(encrypted_data), 'utf-8')
        encrypted_data, equal_signs_count = self.__strip_equal_signs(encrypted_data)
        encrypted_data = iv_start + encrypted_data + iv_end + '='*equal_signs_count # This might result in no equal signs appended.
        return encrypted_data

    def decrypt(self) -> str:
        if not self.__validate_data():
            return None
        iv, encrypted_data = self.__extract_iv()
        if not iv or not encrypted_data:
            return None
        encrypted_data = base64_decode(encrypted_data)
        key = str.encode(self.key)
        iv = str.encode(iv)
        cipher = Cipher(algorithms.AES(key), modes.CBC(iv))
        decryptor = cipher.decryptor()
        decrypted_padded_data = decryptor.update(encrypted_data) + decryptor.finalize()
        unpadder = padding.PKCS7(128).unpadder()
        decrypted_data = unpadder.update(decrypted_padded_data) + unpadder.finalize()
        return str(decrypted_data, 'utf-8')

    def __extract_iv_start_length(self: str) -> tuple[int, str]:
        first_char = self.data[:1]
        if first_char not in self.IV_START_VALUES:
            return None, None
        encrypted_data = self.data[1:]
        iv_start_length = self.IV_START_VALUES.index(first_char) + 1
        return iv_start_length, encrypted_data

    def __extract_iv(self) -> tuple[str, str]:
        iv_start_length, encrypted_data = self.__extract_iv_start_length()
        if not iv_start_length or not encrypted_data:
            return None, None
        iv_end_length = self.IV_LENGTH - iv_start_length
        encrypted_data, equal_signs_count = self.__strip_equal_signs(encrypted_data)
        iv = encrypted_data[:iv_start_length]
        encrypted_data = encrypted_data[iv_start_length:]
        if iv_end_length > 0:
            iv += encrypted_data[iv_end_length*-1:]
        encrypted_data_length = len(encrypted_data)
        encrypted_data = encrypted_data[:encrypted_data_length - iv_end_length]
        encrypted_data += '='*equal_signs_count # This might result in no equal signs appended.
        if len(iv) != self.IV_LENGTH:
            return None, None
        return iv, encrypted_data

    def __strip_equal_signs(self, encrypted_data: str) -> tuple[str, int]:
        equal_signs_count = 0
        encrypted_data_length = len(encrypted_data)
        while encrypted_data[encrypted_data_length - 1] == '=':
            encrypted_data = encrypted_data[:encrypted_data_length - 1]
            encrypted_data_length = len(encrypted_data)
            equal_signs_count += 1
        return encrypted_data, equal_signs_count

    def __validate_data(self) -> bool:
        if not self.key or not self.data:
            return False
        return len(self.data) > self.IV_LENGTH

    def __get_iv_and_start_and_end(self) -> tuple[str, str, str]:
        iv = self.__get_random_iv()
        start_length = self.__get_random_iv_start_length()
        start = self.IV_START_VALUES[start_length - 1] + iv[:start_length]
        end = iv[start_length:]
        return iv, start, end

    def __get_random_iv(self) -> str:
        iv = base64_encode(os_urandom(24)).decode('utf-8')
        for search_value in ['+', '/']:
            iv = iv.replace(search_value, '')
        iv = iv[:self.IV_LENGTH]
        return iv if len(iv) == self.IV_LENGTH else None

    def __get_random_iv_start_length(self):
        start_value = random_choice(self.IV_START_VALUES)
        return self.IV_START_VALUES.index(start_value) + 1


class GrantType:
    AUTH_CODE = 'authorisation_code'
    IMPLICIT = 'implicit';
    USER_CREDENTIAL = 'password'
    CLIENT_CREDENTIALS = 'client_credentials'
    TOKEN_REFRESH = 'refresh_token'
    JWT_BEARER = 'urn:ietf:params:oauth:grant-type:jwt-bearer'


from datetime import datetime, timezone, timedelta
from requests import post as requests_post, codes as requests_status_codes
from bc_time.requests.base import Base as RequestsBase
from bc_time.oauth2.constants.grant_type import GrantType as OAuth2GrantType
from bc_time.api.constants.api import Api as ApiConstants
from bc_time.system.encryption.crypt import Crypt
from bc_time.system.constants.datetime.format import Format as DateTimeFormat

class Token(RequestsBase):
    grant_type = None
    client_id = None
    client_secret = None
    code = None
    crypt_key = None
    private_key_file_path = None
    token = None
    token_expire_as_str = None # This will be stored in UTC.

    __crypt = None

    @property
    def crypt(self) -> Crypt:
        if not self.__crypt:
            self.__crypt = Crypt(key=self.crypt_key)
        return self.__crypt

    @crypt.setter
    def crypt(self, value: Crypt):
        self.__crypt = value

    def __init__(self, client_id: str=None, client_secret: str=None, crypt_key: str=None, grant_type: str=None, code: str=None) -> None:
        self.client_id = client_id
        self.client_secret = client_secret
        self.crypt_key = crypt_key
        self.grant_type = grant_type
        self.code = code

    def request_token(self) -> tuple[bool, dict]:
        if self.__has_valid_token():
            return True, None
        if not self.__validate_data():
            return False, None
        data = self.__get_data()
        if not data:
            return False, None
        request_response = requests_post(
            url=ApiConstants.OAUTH2_TOKEN_URL,
            data=data
        )
        if request_response.status_code != requests_status_codes.ok:
            return False, None
        response_data = self._get_response_data(request_response.text)
        if not set(('access_token', 'expires_in')).issubset(response_data.keys()):
            return False, response_data
        self.token = response_data['access_token']
        token_expire = datetime.now(timezone.utc) + timedelta(seconds=response_data['expires_in'])
        self.token_expire_as_str = token_expire.strftime(DateTimeFormat.MY_SQL_DATE_TIME)
        return True, response_data

    def __has_valid_token(self) -> bool:
        if not self.token or not self.token_expire_as_str:
            return False
        now = datetime.now(timezone.utc)
        return self.token_expire_as_str > now.strftime(DateTimeFormat.MY_SQL_DATE_TIME)

    def __validate_data(self) -> bool:
        if self.grant_type == OAuth2GrantType.AUTH_CODE:
            return self.client_id and self.client_secret and self.code
        elif self.grant_type == OAuth2GrantType.CLIENT_CREDENTIALS:
            return self.client_id and self.client_secret
        elif self.grant_type == OAuth2GrantType.JWT_BEARER:
            return self.client_id and self.private_key_file_path
        return False

    def __get_data(self) -> str:
        if self.grant_type == OAuth2GrantType.AUTH_CODE:
            return {
                'grant_type' : self.grant_type,
                'client_id' : self.client_id,
                'client_secret' : self.__get_client_secret(),
                'code' : self.code,
            }
        elif self.grant_type == OAuth2GrantType.CLIENT_CREDENTIALS:
            return {
                'grant_type' : self.grant_type,
                'client_id' : self.client_id,
                'client_secret' : self.__get_client_secret(),
            }
        elif self.grant_type == OAuth2GrantType.JWT_BEARER:
            return None
        return None

    def __get_client_secret(self):
        if self.crypt_key:
            self.crypt.data = self.client_secret
            return self.crypt.encrypt()
        return self.client_secret


from bc_time.api.objects.base import Base
from bc_time.api.constants.api import Api as ApiConstants

class GroupBase(Base):
    membership_content_type_id = None

    def add_visitor_to_group(self, group_uid: int, visitor_uid: int, add_only: bool=True) -> dict:
        payload = {'link_uid' : visitor_uid}
        if add_only:
            payload['add_only'] = 'add_only'
        return self.api.update(self.membership_content_type_id, group_uid, payload)

    def remove_visitor_from_group(self, group_uid: int, visitor_uid: int, remove_only: bool=True) -> dict:
        payload = {'link_uid' : visitor_uid}
        if remove_only:
            payload['remove_only'] = 'remove_only'
        return self.api.update(self.membership_content_type_id, group_uid, payload)

    def get_all_members_pagination(self, group_uid: int, page: int=1, row_count: int=ApiConstants.DEFAULT_ROW_COUNT) -> dict:
        return self.api.get_all_using_pagination(self.membership_content_type_id, content_uid=group_uid, page=page, row_count=row_count)


from json import loads as json_loads

class Base:
    def _get_response_data(self, response_text: str) -> dict:
        response_data = None
        if self.crypt_key:
            self.crypt.data = response_text
            response_data = self.crypt.decrypt()
        if not response_data:
            response_data = response_text
        try:
            return json_loads(response_data)
        except ValueError as exception:
            print(f"json.loads: {str(exception)}")
        return None


from enum import unique, IntEnum

@unique
class ContentType(IntEnum):
    user = 10
    device = 18
    employee = 22
    branch = 35
    employee_leave = 50
    settings = 86
    employee_overtime_rule_detail = 103
    employee_group_membership = 105
    department = 39
    visitor = 95
    visitor_group = 99
    security_details = 250
    security_area_duration = 252 # How much time spent in area(s).
    visitor_security_details = 350
    visitor_security_area_duration = 352 # How much time spent in area(s).
    controller = 111
    company_profile_user_membership = 114
    company_profile_user_membership_email_confirm = 115
    visitor_group_membership = 116
    visitor_group_area_membership = 117
    current_company_profile = 118


class Validate:
    @staticmethod
    def is_numeric(value, min=None, max=None):
        if not str(value).isnumeric():
            return False
        is_min_numeric = str(min).isnumeric()
        is_max_numeric = str(max).isnumeric()
        if is_min_numeric and is_max_numeric:
            return value >= min and value <= max
        elif is_min_numeric:
            return value >= min
        elif is_max_numeric:
            return value <= max
        return True


from bc_time.api.objects.base import Base
from bc_time.api.constants.api import Api as ApiConstants

class ObjectBase(Base):
    def create(self, payload: dict) -> dict:
        return self.api.create( content_type_id=self.content_type_id, payload=payload)

    def create_many(self, payloads: dict) -> dict:
        return self.api.create(
            content_type_id=self.content_type_id,
            payloads=payloads,
            content_uid=ApiConstants.UID_POST_MANY
        )

    def update(self, content_uid, payload: dict) -> dict:
        return self.api.update(
            content_type_id=self.content_type_id,
            content_uid=content_uid,
            payload=payload
        )

    def update_many(self, payloads: dict) -> dict:
        return self.api.update(
            content_type_id=self.content_type_id,
            content_uid=ApiConstants.UID_POST_MANY,
            payload=payloads
        )

    def get_all_using_pagination(self, filters: dict=None, page: int=1, row_count: int=ApiConstants.DEFAULT_ROW_COUNT) -> dict:
        return self.api.get_all_using_pagination(
            content_type_id=self.content_type_id,
            filters=filters,
            page=page,
            row_count=row_count
        )

    def get_one(self, content_uid: int) -> dict:
        return self.api.get_one(content_type_id=self.content_type_id, content_uid=content_uid)

    def get_many(self, content_uids: list) -> dict:
        return self.api.get_many(content_type_id=self.content_type_id, content_uids=content_uids)


class Api:
    DOMAIN = 'http://time-v2.local'
    OAUTH2_TOKEN_URL = 'http://time-v2.local/oauth2/token/'
    OAUTH2_AUTHORISE_URL = 'http://time-v2.local/oauth2/authorise/'
    API_URL = 'http://time-v2.local/api/'
    DEFAULT_ROW_COUNT = 100 # 100 is the maximum no. of rows the API allows to retrieved at a time.
    UID_GET_ALL = -1
    UID_POST_MANY = -2


from os.path import exists as path_exists
from pathlib import Path
from json import dumps as json_dumps
from requests import post as requests_post, get as requests_get, codes as requests_status_codes
from configparser import ConfigParser
from bc_time.system.encryption.crypt import Crypt
from bc_time.system.validate import Validate
from bc_time.requests.base import Base as RequestsBase
from bc_time.oauth2.constants.grant_type import GrantType as OAuth2GrantType
from bc_time.oauth2.token import Token
from bc_time.api.constants.api import Api as ApiConstants
from bc_time.api.enumerators.content_type import ContentType

class Api(RequestsBase):
    client_id = None
    client_secret = None
    crypt_key = None
    grant_type = None
    code = None

    __crypt = None
    __token = None

    @property
    def crypt(self) -> Crypt:
        if not self.__crypt:
            self.__crypt = Crypt(key=self.crypt_key)
        return self.__crypt

    @property
    def token(self):
        if not self.__token:
            self.__token = Token(
                client_id=self.client_id,
                client_secret=self.client_secret,
                crypt_key=self.crypt_key,
                grant_type=self.grant_type,
                code=self.code
            )
        return self.__token

    def __init__(self, grant_type: str=None, code: str=None) -> None:
        self.__init_authentication_credentials_from_file()
        if grant_type:
            self.grant_type = grant_type
        if code:
            self.code = code
        self.token.crypt = self.crypt

    def __init_authentication_credentials_from_file(self, file_path: str='.bc_time/credentials', section: str='default'):
        time_config_file_path = f"{str(Path.home())}/{file_path}"
        if not path_exists(time_config_file_path):
            return
        config_parser = ConfigParser(inline_comment_prefixes=';')
        config_parser.read(time_config_file_path)
        if section not in config_parser:
            return
        config_data_keys_and_attributes = ['client_id', 'client_secret', 'crypt_key', 'grant_type']
        for config_data_key_or_attribute in config_data_keys_and_attributes:
            if config_data_key_or_attribute in config_parser[section]:
                setattr(
                    self,
                    config_data_key_or_attribute,
                    config_parser[section][config_data_key_or_attribute]
                )

    def create(self, content_type_id: ContentType, payload: dict, content_uid: int=None) -> dict:
        request_token_result, request_token_response_data = self.token.request_token()
        if not request_token_result:
            return request_token_response_data
        create_payload = self.__get_create_or_update_data(
            content_type_id=content_type_id,
            payload=payload,
            content_uid=content_uid
        )
        post_response = requests_post(
            url=ApiConstants.API_URL,
            data=create_payload
        )
        return self._get_response_data(post_response.text) if post_response.status_code == requests_status_codes.ok else None

    def update(self, content_type_id: ContentType, content_uid: int, payload: dict) -> dict:
        request_token_result, request_token_response_data = self.token.request_token()
        if not request_token_result:
            return request_token_response_data
        update_payload = self.__get_create_or_update_data(
            content_type_id,
            content_uid=content_uid,
            payload=payload
        )
        post_response = requests_post(
            url=ApiConstants.API_URL,
            data=update_payload
        )
        return self._get_response_data(post_response.text) if post_response.status_code == requests_status_codes.ok else None

    def __get_create_or_update_data(self, content_type_id: ContentType, payload: dict, content_uid: int=None) -> dict:
        data = {'content_type_id' : int(content_type_id)}
        if content_uid: # Will be omitted if performing POST (1 new object).
            data['content_uid'] = content_uid
        if payload:
            if content_uid != ApiConstants.UID_POST_MANY:
                data.update(payload)
            else:
                data['data'] = payload
        create_or_update_payload = {'access_token' : self.token.token}
        if self.crypt_key:
            self.crypt.data = json_dumps(data)
            create_or_update_payload['data'] = self.crypt.encrypt()
        else:
            create_or_update_payload.update(data)
        return create_or_update_payload

    def get_all_using_pagination(self, content_type_id: ContentType, content_uid: int=ApiConstants.UID_GET_ALL, filters: dict=None, page: int=1, row_count: int=ApiConstants.DEFAULT_ROW_COUNT) -> dict:
        if not self.token.request_token():
            return None
        request_params = self.__get_request_params(
            content_type_id=content_type_id,
            content_uids=[content_uid],
            filters=filters,
            page=page,
            row_count=row_count
        )
        request_response = requests_get(
            url=ApiConstants.API_URL,
            params=request_params
        )
        return self._get_response_data(request_response.text) if request_response.status_code == requests_status_codes.ok else None

    def get_one(self, content_type_id: ContentType, content_uid: int) -> dict:
        if not self.token.request_token():
            return None
        request_params = self.__get_request_params(content_type_id, content_uids=[content_uid])
        request_response = requests_get(
            url=ApiConstants.API_URL,
            params=request_params
        )
        return self._get_response_data(request_response.text) if request_response.status_code == requests_status_codes.ok else None

    def get_many(self, content_type_id: ContentType, content_uids: list) -> dict:
        if not self.token.request_token():
            return None
        request_params = self.__get_request_params(content_type_id, content_uids)
        request_response = requests_get(
            url=ApiConstants.API_URL,
            params=request_params
        )
        return self._get_response_data(request_response.text) if request_response.status_code == requests_status_codes.ok else None

    def __get_request_params(self, content_type_id: ContentType, content_uids: list=None, filters: dict=None, page: int=None, row_count: int=None) -> dict:
        data = {
            'content_type_id' : int(content_type_id),
            'content_uid' : content_uids[0] if len(content_uids) == 1 else ','.join([str(content_uid) for content_uid in content_uids]),
        }
        if self.__can_paginate(page, row_count):
            data['page'] = page
            data['row_count'] = row_count
        if filters:
            data.update(filters)
        request_params = {'access_token' : self.token.token}
        if self.crypt_key:
            self.crypt.data = json_dumps(data)
            request_params['data'] = self.crypt.encrypt()
        else:
            request_params.update(data)
        return request_params

    def __can_paginate(self, page: int, row_count: int) -> bool:
        if not Validate.is_numeric(page, min=1):
            return False
        return Validate.is_numeric(row_count, min=1, max=ApiConstants.DEFAULT_ROW_COUNT)


from enum import unique, IntEnum

@unique
class RequestStatus(IntEnum):
    none = 0
    success = 1
    data_invalid = 2
    insufficient_access = 3
    error = 4
    client_id_invalid = 5
    client_secret_invalid = 6
    method_invalid = 7
    content_type_id_invalid = 8
    uid_invalid = 9
    field_invalid = 10
    user_uid_invalid = 11
    post_batch_data_invalid = 12
    data_empty = 13
    no_response = 14 # No internet connection, blocked by firewall, etc.; used exclusively by Time Comm.
    response_invalid = 15 # Http code is anything other than 200 (OK); used exclusively by Time Comm.
    response_json_invalid = 16 # Used exclusively by Time Comm.
    company_profile_inactive = 17
    cost_plan_overdue = 18
    token_invalid = 19
    token_expired = 20
    authentication_credentials_invalid = 21 # Raised by Time Comm. Service.